package br.pucpr.neuralnetwork;

// Classe que representa um neurnio
public class Neuron {
	// Dendritos
	private Dendrite[] dendrites;
	// Axnio
	private Axon axon;
	// Pesos
	private float[] weights;
	// Pesos padres
	private float[] defaultWeights;
	// Tipo da funo de ativao
	private String activationType;
	// Constante de aprendizado
	private float constantLearning;

	// Mtodo construtor
	// aceita como prametros:
	// * numberDendrites = Nmero de dendritos
	// * defaultWeights = Pesos padro
	// * activationType = tipo da funo de ativao
	public Neuron(int numberDendrites, float[] defaultWeights, String activationType) {
		// Define os pesos padres
		this.defaultWeights = defaultWeights;
		
		// Se no for definida o tipo da funo de ativao,
		// ento define como sign
		if (activationType == null) {
			this.activationType = "sign";
		}
		
		// Pesos
		weights = new float[numberDendrites];
		// Dendritos
		dendrites = new Dendrite[numberDendrites];
		
		// Cria valores iniciais randmicos para os pesos
		for (int i = 0; i < weights.length; i++) {
			weights[i] = Util.random(-1, 1);
		}
		
		// Constante de aprendizado
		constantLearning = (float)0.0001;
		// Axnio
		axon = new Axon();
	}
	
	// Soma os dendritos com os seus pesos
	// mdia ponderada
	public float sumDendrites() {
		float sum = 0;
		
		if (defaultWeights != null && defaultWeights.length > 0) {
			for (int j = 0; j < defaultWeights.length; j++) {
				weights[j] = defaultWeights[j];
			}
		}
		
		for (int i = 0; i < weights.length; i++) {
			sum += dendrites[i].getValue() * weights[i];
		}
		return sum;
	}
	
	// Mtodo de ativao
	// dependendo da soma e do tipo de ativao
	// envia o sinal -1 ou 1 para o axnio
	public void activation() {
		float sum = sumDendrites();

		axon.setSign(1);
		if (sum < 0) {
			axon.setSign(-1);
		}
	}
	
	// Treina o neurnio, afim de obter os melhores pesos para os dendritos
	// aceita como parmetros:
	// * numberTimes = nmero de pocas
	// * percentageCorrect = percentual correto para o treinamento ser considerado bem sucedido
	// * examplesTraining = exemplos para o treinamento
	public void train(int numberTimes, int percentageCorrect, Point[] examplesTraining) {
		System.out.println("Treinamento iniciado...");
		
		// Quantidade de treinamento verificado
		int totalVerified = 0;
		// Quantidade de treinamento verificado errado
		int totalVerifiedWrong = 0;
		// Quantidade de treinamento verificado correto
		int totalVerifiedCorrect = 0;
		
		// Repete at o nmero de pocas ser satisfeito
		for (int i = 0; i < numberTimes; i++) {
			// Percorre todos os exemplos de treinamento
			for (int j = 0; j < examplesTraining.length; j++) {
				// Obtm a sada adivinhada pelo neurnio
				int result = guess(examplesTraining[j]);
				// Calcula o fator de mudana do peso baseado no erro
				// erro = sada desejada - sada adivinhada
				// multiplica pela constante de aprendizado
				float weightChange = constantLearning * (examplesTraining[j].getOutput() - result);
				// Ajusta os pesos baseado no fator de mudana * a entrada
				for (int k = 0; k < weights.length; k++) {
					weights[k] += weightChange * examplesTraining[j].getVals()[k];
				}
				
				// Contabiliza se o valor adivinhado foi correto ou no
				if (result == examplesTraining[j].getOutput()) {
					totalVerifiedCorrect++;
				} else {
					totalVerifiedWrong++;
				}
			}
		}
		
		// Quantidade de treinamentos verificados
		totalVerified = totalVerifiedWrong + totalVerifiedCorrect;
		float totalPercentageCorrect = ((100 * totalVerifiedCorrect ) / totalVerified);
		
		System.out.println("Treinamento terminado...");
		System.out.println("Resultados obtidos:");
		System.out.println(" - Nmero de pocas executadas: " + numberTimes);
		
		String indicativeSuccess = "bem sucedido";
		
		if (totalPercentageCorrect < percentageCorrect) {
			indicativeSuccess = "mal sucedido";
		}
		
		System.out.println(" - Indicativo de sucesso: Treinamento " + indicativeSuccess);
		System.out.println(" - Porcentagem correto: " + totalPercentageCorrect);
		System.out.println(" - Pesos:");
		
		if (defaultWeights != null && defaultWeights.length > 0) {
			for (int j = 0; j < defaultWeights.length; j++) {
				weights[j] = defaultWeights[j];
			}
		}
		
		for (int i = 0; i < weights.length; i++) {
			System.out.println("   - Peso " + i + ": " + weights[i]);	
		}
	}
	
	// Mtodo que adivinha a sada esperada
	// baseada nas entradas, de acordo com os pesos
	// calculados
	public int guess(Point point) {
		// Define os valores para os dendritos
		setDendrites(point.getVals());
		// Faz a ativao
		activation();
		// Sinal de ativao
		return axon.getSign();
	}

	// Define os valores para os dendritos
	// de acordo com um array de valores
	public void setDendrites(int[] vals) {
		for (int i = 0; i < vals.length; i++) {
			dendrites[i] = new Dendrite(vals[i]);
		}
	}

	// Obtm a constante de aprendizado
	public float getConstantLearning() {
		return constantLearning;
	}

	// Define uma nova constante de aprendizado
	public void setConstantLearning(float constantLearning) {
		this.constantLearning = constantLearning;
	}

	// Obtm o axnio
	public Axon getAxon() {
		return axon;
	}

	// Obtm o tipo da funo de ativao usada
	public String getActivationType() {
		return activationType;
	}
}
